import subprocess, functools, os, re, datetime, json
from subprocess import run, DEVNULL
import types, pyarrow.parquet as pq, re
from pathlib import Path
from .config import bblib, token
from logger import logger
import os, urllib.request as request, functools, dvc.api
from urllib.parse import urlparse

from .checks import check_url_available, check_token, check_symlink_permission

class Brick:

    def __init__(self, remote, commit):
        self.remote = remote
        self.commit = commit
    
    @staticmethod
    def FromURL(url):
        """Get the version of a brick from its git repo."""
        remote, commit = url.split("#")
        return Brick(remote, commit)

    @staticmethod
    def FromPath(path):
        """Get the version of a brick from its git repo."""
        bdir = bblib(path)
        gsys = functools.partial(subprocess.check_output, shell=True, cwd=bdir)
        commit = gsys("git rev-parse HEAD").decode().strip()
        remote = gsys("git config --get remote.origin.url").decode().strip()
        return Brick(remote, commit)

    @staticmethod
    def FromRemote(remote):
        """Get the version of a brick from its git repo."""
        try:
            logger.info(f"getting latest version of {remote}")
            commit = subprocess.check_output(f'git ls-remote "{remote}" HEAD', shell=True)
            commit = commit.decode().strip().split()[0]
            return Brick(remote, commit)
        except subprocess.CalledProcessError as e:
            logger.error(f"failed to get latest version of {remote}: {e}")
            logger.error(f"is {remote} a valid git repository?")
            return None

    @staticmethod
    def Resolve(ref, force_remote=False):
        """find all bricks matching ref. `ref` can be:
            - existing  name ie. 'tox21'
            - git-url syntax ie. 'https://github.com/biobricks-ai/tox21#commit'
        if `force_remote` is True then retrieve brick from remote repository"""
        # TODO this should resolve from the .bb directory when in a biobrick repo

        # if name matches remote#commit then resolve from url        
        if re.match("^http.*[0-9a-f]{40}$",ref):
            return Brick.FromURL(ref)
        
        # if name matches remote then resolve from remote
        if re.match("^http.*$",ref):
            return Brick.FromRemote(ref)
        
        # otherwise resolve to https://github.com/biobricks-ai/<name>
        remote = f"https://github.com/biobricks-ai/{ref}"

        if force_remote: 
            return Brick.FromRemote(remote)

        # retrieve from library if it exists
        bricks = []
        bdir = bblib() / "biobricks-ai" / ref
        if bdir.exists():
            for bdir in (bblib() / "biobricks-ai" / ref).iterdir():
                logger.info(f"checking {bdir.name} for {ref}")
                # if the directory is a sha hash then add it to brick array
                if bdir.is_dir() and re.match("[0-9a-f]{40}$",bdir.name):
                    brick = Brick.FromPath(bdir)
                    if remote == brick.remote:
                        bricks.append(brick)
        
        # sort the bricks by their commit_date
        bricks.sort(key=lambda b: b.get_commit_date(), reverse=True)
        
        # pick the most recent one
        if len(bricks) > 0:
            return bricks[0]
        
        # if we can't find it in the library, then get the remote version
        return Brick.FromRemote(remote)


    def get_commit_date(self):
        """get the date of the commit"""
        gsys = functools.partial(subprocess.check_output, shell=True, cwd=self.path())
        date = gsys("git show -s --format=%ci").decode().strip()
        return datetime.datetime.strptime(date, "%Y-%m-%d %H:%M:%S %z")

    def url(self):
        return f"{self.remote}#{self.commit}"
    
    def urlpath(self):
        return Path(urlparse(self.url()).path[1:])

    def path(self):
        return bblib() / self.urlpath() / self.commit

    def _relpath(self):
        "get the path to this brick relative to bblib"
        return self.urlpath() / self.commit

    def install(self):
        "install this brick"
        logger.info(f"running checks on brick")
        check_url_available(self.remote)
        check_token(token())
        check_symlink_permission()

        if bblib(self.commit).exists():
            logger.info(f"\033[91m{self.url}\033[0m already exists in BioBricks library.")
            return True

        cmd = functools.partial(run,shell=True,stdout=DEVNULL,stderr=DEVNULL)
        
        # old way - cmd(f"git submodule add {self.remote} {self.repo}",cwd=bblib())
        logger.info(f"git clone {self.remote} {self._relpath()} in {bblib()}")
        cmd(f"git clone {self.remote} {self._relpath()}", cwd = bblib())
        cmd(f"git checkout {self.commit}", cwd = self.path())

        logger.info(f"adding brick to dvc cache")
        rsys = functools.partial(cmd,cwd=self.path())
        rsys(f"dvc cache dir {bblib() / 'cache'}")
        rsys("dvc config cache.shared group")
        rsys("dvc config cache.type symlink")

        # SET UP BIOBRICKS.AI DVC REMOTE WITH AUTH
        logger.info(f"setting up credentials for dvc.biobricks.ai")
        rsys("dvc remote add -f biobricks.ai https://dvc.biobricks.ai")
        rsys("dvc remote modify --local biobricks.ai auth custom")
        rsys("dvc remote modify --local biobricks.ai custom_auth_header BBToken")
        rsys("dvc remote modify --local biobricks.ai read_timeout 300")
        rsys("dvc remote modify --local biobricks.ai connect_timeout 300")

        rsys(f"dvc remote modify --local biobricks.ai password {token()}")

        logger.info(f"discovering brick assets dvc.biobricks.ai")
        fs = dvc.api.DVCFileSystem(self.path())
        paths = fs.find("data",maxdepth=1) + fs.find("brick",maxdepth=1)
        parquet_paths = [x for x in paths if x.endswith('.parquet')]

        logger.info(f"pulling brick assets")
        run(f"dvc pull {' '.join(parquet_paths)}", cwd=self.path(), shell=True)
        
        logger.info(f"\033[94m{self.url()}\033[0m succesfully downloaded to BioBricks library.")
        return self
    
    def load(self):
        "load this brick"
        bdir = self.path()
        if not bdir.exists(): 
            raise Exception(f"no path '{bdir}' try `biobricks install {self.url()}`")
        
        def dirns(dir: Path):
            filter = lambda d: d.name.endswith('.parquet')
            paths = [d for d in dir.rglob('*') if filter(d)]
            namespace = types.SimpleNamespace()
            pkey = lambda p: re.sub(r'[.-]','_',p)
            for p in paths:
                path = p.relative_to(dir)
                logger.info(f"loading {path}...")
                current = namespace
                for part in path.parts:
                    key = pkey(part)
                    if not hasattr(current, key) and part.endswith('.parquet'):
                        setattr(current, key[:-8], pq.ParquetFile(str(p)))
                    elif not hasattr(current, key):
                        setattr(current, key, types.SimpleNamespace())
                        current = getattr(current, key)
                    elif hasattr(current, key):
                        current = getattr(current, key)
            logger.info(f"loaded {len(paths)} tables from {dir}")
            return namespace

        ns1 = dirns(bdir / 'data')
        ns2 = dirns(bdir / 'brick')
        result = {**ns2.__dict__, **ns1.__dict__}
        return types.SimpleNamespace(**result)
    
    def assets(self):
        "get the assets for this brick"
        bdir = self.path()
        if not bdir.exists(): 
            raise Exception(f"no path '{bdir}' try `biobricks install {self.url()}`")
        
        def find_parquet_files(directory):
            if not directory.exists():
                return []
            for entry in os.scandir(directory):
                if entry.name.endswith('.parquet'):
                    yield entry.path
                elif entry.is_dir():
                    yield from find_parquet_files(entry.path)
                
        return list(find_parquet_files(bdir / 'data')) + list(find_parquet_files(bdir / 'brick'))
    
    def uninstall(self):
        "uninstall this brick"
        os.rmdir(self.path())
        


import urllib.request as request
import json, urllib
import os, tempfile, uuid

def check_url_available(url):
    try:
        code = request.urlopen(url).getcode()
        if code!=200:
            raise Exception(f"{url} not available")
    except Exception as e:
        raise Exception(f"{url} not available") from e

def check_token(token, silent=False):
    """verify that the token is a valid biobricks.ai token"""
    url = f"https://biobricks.ai/token/is_valid?token={token}"
    is_valid = json.loads(urllib.request.urlopen(url).read())
    if not is_valid and not silent: 
        raise ValueError(f"Invalid token. Run `biobricks configure`")
    return is_valid

def check_symlink_permission():
    try:
        src = tempfile.NamedTemporaryFile()
        dst = os.path.join(tempfile.gettempdir(), str(uuid.uuid4()))
        os.symlink(src.name, dst)
        os.remove(dst)
        src.close()
        return True
    except Exception as e:
        raise PermissionError("Need Symlink Permission. Contact administrator and see https://dvc.org/doc/user-guide/how-to/run-dvc-on-windows#enable-symbolic-links.")

def check_configured():
    """check that biobricks is configured"""
    try:
        from .config import read_config, check_has_bblib
        if not read_config(): raise Exception("no config")
        if not check_has_bblib(): raise Exception("no bblib")
        if not check_token(read_config()["TOKEN"], silent=True): raise Exception("invalid token")
        return True
    except Exception as e:
        raise Exception("biobricks not configured. run `biobricks configure`") from e


import sys
import os
import biobricks as bb
import click, cloup
from logger import logger
from pathlib import Path
import pathlib as pl, yaml, pkg_resources
from .config import read_config, write_config, init_bblib
from .checks import check_token
from .brick import Brick

@cloup.group('biobricks')
def cli():
    pass

class Sect:
    GLOBAL = cloup.Section('GLOBAL: modify global config and installed bricks')
    BRICK = cloup.Section('LOCAL: build new bricks and manage their dependencies')

@cli.command(help="configure brick path and token",section=Sect.GLOBAL)
@click.option("--bblib", default=None, type=click.Path(), help="path to store bricks")
@click.option("--token", default=None, help="biobricks.ai/token auth token")
@click.option("--overwrite", default=False, help="overwrite existing config?")
def configure(bblib, token, overwrite):

    path = Path.home().joinpath(".biobricks")
    config = read_config()

    # CHECK IF CONFIG WILL OVERWRITE EXISTING
    msg = click.style("WARNING: overwrite existing config?", fg="red")
    if path.exists() and not overwrite and not click.confirm(msg):
        sys.exit(0)
    
    # BBLIB PROMPT - DEFAULT TO EXISTING
    conmsg = lambda: f"use current BBLIB '{bblib or config['BBLIB']}'?"
    if not bblib and config.keys() >= {"BBLIB"} and click.confirm(conmsg(), default=True):
        bblib = config["BBLIB"]
    elif not bblib:
        bblib = click.prompt("Choose path to store bricks", type=click.Path())

    # TOKEN PROMPT - DEFAULT TO EXISTING AND THEN FREE TOKEN
    deftoken = "skip for temporary token"
    conmsg = f"use current token defined at '{path}'?"
    if not token and config.keys() >= {"TOKEN"} and click.confirm(conmsg):
        token = config["TOKEN"]
    elif not token:
        msg = "Copy token from biobricks.ai/token"
        token = click.prompt(msg, hide_input=True, default = deftoken)

    # VALIDATE TOKEN    
    while token != deftoken and not check_token(token, silent=True):
        click.echo(click.style("invalid token. check your token at https://biobricks.ai/token", fg="red"))
        token = click.prompt("Input a token from biobricks.ai/token",hide_input=True, default=deftoken)
    token = "VQF6Q2U-NKktZ31ioVYa9w" if token == deftoken else token

    # write configuration
    config = { "BBLIB": bblib, "TOKEN": token }
    write_config(config)

    # initialize bblib
    bblib = init_bblib()

    msg = f"Done! BioBricks has BBLIB {bblib} and config {path}"
    click.echo(click.style(msg, fg="green"))


@cli.command(help="Initialize a .bb directory for data dependencies",
             section=Sect.BRICK)
def init():
    location = ".bb"
    dotbb = pl.Path(location)
    if dotbb.exists():
        return
    dotbb.mkdir()
    
    with open(dotbb / ".gitignore", "w") as f: 
        f.write("/*/") # ignore all subdirectories

def local_bblib():
    return Path(".bb")

def check_has_local_bblib():
    if not local_bblib().exists():
        raise Exception(".bb not found. run `biobricks init` first.")

def symlink_local_brick(brick):
    check_has_local_bblib()
    localpath = local_bblib() / brick.urlpath()
    localpath.parent.mkdir(parents=True, exist_ok=True)
    os.symlink(brick.path(), localpath)
    
    # write a line to the dependencies file recording this import
    with open(local_bblib() / "dependencies.txt", "a") as f:
        f.write(f"{brick.url()}\n")

@cli.command(name="add",
    help="Import a data dependency into the .bb directory",
    section=Sect.BRICK)
@click.argument("ref",type=str)
def add(ref):
    check_has_local_bblib()
    brick : Brick = Brick.Resolve(ref, force_remote=True).install()
    symlink_local_brick(brick)
    
@cli.command(name="pull", help="install all the local dependencies",
    section=Sect.BRICK)
def pull():
    check_has_local_bblib()
    with open(local_bblib() / "dependencies.txt", "r") as f:
        for line in f.readlines():
            Brick.FromURL(line).install()

@cli.command(help="Install a data dependency into $BBLIB",
    section=Sect.GLOBAL)
@click.argument("ref",type=str)
def install(ref):
    return Brick.Resolve(ref, force_remote=True).install()

@cli.command(help="Show the status of the local brick",
    section=Sect.BRICK)
def status():
    print("BBLIB: " + str(bb.bblib()))
    # print the dependencies file
    with open(local_bblib() / "dependencies.txt", "r") as f:
        print(f.read())

if __name__ == "__main__":
    cli()



import pytest
import os
import pathlib
import tempfile

import biobricks as bb

@pytest.fixture
def CONFIG():
    oldbb = os.getenv('BBLIB')
    tdir  = tempfile.TemporaryDirectory()
    os.environ['BBLIB'] = tdir.name
    yield pathlib.Path(tdir.name)
    if oldbb: os.environ['BBLIB'] = oldbb
    tdir.cleanup()

# def test_initialize(BBLIB,TOKEN):
#     errmsg = "must use token from https://biobricks.ai/token"
#     with pytest.raises(Exception,match=errmsg) as e:
#         bb.initialize()

#     res = bb.initialize(TOKEN)
#     assert(res == BBLIB)
#     assert(bb.token() == TOKEN)

# def test_bblib(EMPTY_BBLIB):
#     errmsg = "'BBLIB' env not available.*"
#     with pytest.raises(Exception,match=errmsg):
#         bb.bblib()

# def test_load(local_bblib):
#     brick="lkjasdfkjasdklfj"
#     with pytest.raises(Exception,match=".*not available.*"):
#         bb.pull(brick)

#     brick = "hello-brick"
#     org = "biobricks-ai"
#     bb.pull(brick,org)
#     assert bb.bblib(f'{org}/{brick}/data/mtcars.parquet').exists()

#     tbls = bb.load(brick,org)
#     assert tbls.mtcars.shape == (32,11)

#     with pytest.raises(Exception,match="no path.*"):
#         bb.load("a-brick-that-doesn't-exist")
    
#     with pytest.raises(Exception,match=".* not available"):
#         check_url_available("http://the-internet.herokuapp.com/status_codes/301")


# In a shell run 
# > pip install biobricks
# > biobricks configure
import biobricks as bb, seaborn as sns, matplotlib.pyplot as plt

# Install and load clinvar
cv = bb.install("clinvar").load() 

# load `variant_summary` and `var_citations` into pandas
variants = cv.variant_summary.to_pandas()
citation = cv.var_citations.to_pandas()
joined = variants.merge(citation, on="AlleleID")

# filter to only the pathogenic variants
pathogenic = joined[joined.ClinicalSignificance == "Pathogenic"]

# Find the top 10 genes by number of publications
counts = pathogenic.groupby(["GeneSymbol"]).size().reset_index(name="count")
counts = counts.sort_values(by="count", ascending=False)[0:10]

# build a barplot and output it to plot.jpeg
ax = sns.barplot(x='GeneSymbol', y='count', data=counts, color='steelblue')
plt.savefig('pathgenes.png',dpi=300, bbox_inches='tight')





from .brick import Brick
from .cli import cli
from .checks import check_configured
import requests, json 

def load(brick):
    """Load a brick from the local filesystem"""
    check_configured()
    return Brick.Resolve(brick).load()

def assets(brick):
    """List the paths of the assets of a brick"""
    check_configured()
    return [str(x) for x in Brick.Resolve(brick).assets()]

def path(brick):
    check_configured()
    return Brick.Resolve(brick).path()

def install(brick):
    """Install a brick from a remote repository"""
    check_configured()
    return Brick.Resolve(brick).install()

def configure():
    """Configure biobricks globally"""
    return cli.config()
    
def ls_remote():
    """List the bricks available on github.com/biobricks-ai"""
    check_configured()
    r = requests.get("https://api.github.com/users/biobricks-ai/repos")
    repos = json.loads(r.text)
    for repo in repos:
        yield repo["name"]


import os, warnings, json, urllib
from subprocess import run, DEVNULL, STDOUT
from pathlib import Path

def read_config():
    path = Path.home().joinpath(".biobricks")
    if not path.exists(): return {}
    return json.loads(path.read_text())

def write_config(config):
    path = Path.home().joinpath(".biobricks")
    path.write_text(json.dumps(config))

def init_bblib() -> None:
    bbpath = Path(read_config()['BBLIB'])
    os.makedirs(bbpath, exist_ok=True)
    os.makedirs(bbpath / "cache", exist_ok=True)
    run("git init", cwd=bbpath, stdout=DEVNULL, stderr=STDOUT, shell=True)
    return bblib()

def check_has_bblib():
    if not 'BBLIB' in read_config(): return False
    if not Path(read_config()["BBLIB"]).exists(): return False
    if not Path(read_config()["BBLIB"]).is_dir(): return False
    if not Path(read_config()["BBLIB"]).joinpath("cache").exists(): return False
    if not Path(read_config()["BBLIB"]).joinpath(".git").exists(): return False
    return True

def bblib(path=""):
    if check_has_bblib(): return Path(read_config()["BBLIB"]) / path
    raise Exception("no BBLIB path. run `biobricks configure` to set your BBLIB path")

def token_url():
    return "https://biobricks.ai/token"

def token():
    if read_config().keys() >= {"TOKEN"}: return read_config()["TOKEN"]
    raise Exception("no token. run `biobricks configure` to set your token")
    
